@@ -133,8 +133,6 @@ static void get_off_job_bar(bg_job_t *job);
 static bg_job_t * add_background_job(pid_t pid, const char cmd[],
 		uintptr_t err, uintptr_t data, BgJobType type, int with_bg_op);
 static void * background_task_bootstrap(void *arg);
-static void set_current_job(bg_job_t *job);
-static void make_current_job_key(void);
 static int update_job_status(bg_job_t *job);
 static void mark_job_finished(bg_job_t *job, int exit_code);
 static int bg_op_cancel(bg_op_t *bg_op);
@@ -148,19 +146,25 @@ static pthread_mutex_t new_err_jobs_lock = PTHREAD_MUTEX_INITIALIZER;
 /* Conditional variable to signal availability of new jobs in new_err_jobs. */
 static pthread_cond_t new_err_jobs_cond = PTHREAD_COND_INITIALIZER;
 
-/* Thread local storage for bg_job_t associated with active thread. */
+/* Thread-local storage for bg_job_t associated with active thread. */
 static pthread_key_t current_job;
 
-void
+int
 bg_init(void)
 {
-	pthread_t id;
-	int err = pthread_create(&id, NULL, &error_thread, NULL);
-	assert(err == 0);
-	(void)err;
+	/* Create thread-local storage before starting any background threads. */
+	if(pthread_key_create(&current_job, NULL) != 0)
+	{
+		return 1;
+	}
 
-	/* Initialize state for the main thread. */
-	set_current_job(NULL);
+	pthread_t id;
+	if(pthread_create(&id, NULL, &error_thread, NULL) != 0)
+	{
+		return 1;
+	}
+
+	return 0;
 }
 
 void
@@ -189,16 +193,30 @@ bg_check(void)
 	{
 		job_check(p);
 
-		pthread_spin_lock(&p->status_lock);
-		int running = p->running;
-		int can_remove = (!p->running && p->use_count == 0);
-		pthread_spin_unlock(&p->status_lock);
+		/* In case of lock failure, assume the job is active. */
+		int running = 1;
+		int can_remove = 0;
 
-		active_jobs += (running != 0 && p->in_menu);
-
-		if(!running && p->on_job_bar)
+		if(pthread_spin_lock(&p->status_lock) == 0)
 		{
-			get_off_job_bar(p);
+			running = p->running;
+			can_remove = (!running && p->use_count == 0);
+			(void)pthread_spin_unlock(&p->status_lock);
+		}
+
+		active_jobs += (running && p->in_menu);
+
+		if(!running)
+		{
+			if(p->on_job_bar)
+			{
+				get_off_job_bar(p);
+			}
+			if(p->exit_cb != NULL)
+			{
+				p->exit_cb(p, p->exit_cb_arg);
+				p->exit_cb = NULL;
+			}
 		}
 
 		/* Remove job if it is finished now. */
@@ -253,11 +271,14 @@ job_check(bg_job_t *job)
 	/* Display portions of errors from the job while there are any. */
 	do
 	{
-		pthread_spin_lock(&job->errors_lock);
-		new_errors = job->new_errors;
-		job->new_errors = NULL;
-		job->new_errors_len = 0U;
-		pthread_spin_unlock(&job->errors_lock);
+		new_errors = NULL;
+		if(pthread_spin_lock(&job->errors_lock) == 0)
+		{
+			new_errors = job->new_errors;
+			job->new_errors = NULL;
+			job->new_errors_len = 0U;
+			(void)pthread_spin_unlock(&job->errors_lock);
+		}
 
 		if(new_errors != NULL && !job->skip_errors)
 		{
@@ -377,7 +398,7 @@ bg_and_wait_for_errors(char cmd[], const struct cancellation_t *cancellation)
 		}
 		else
 		{
-			result = status_to_exit_code(get_proc_exit_status(pid));
+			result = status_to_exit_code(get_proc_exit_status(pid, cancellation));
 		}
 	}
 
@@ -460,11 +481,10 @@ error_thread(void *p)
 				job = &j->err_next;
 			}
 
-			if(!need_update_list)
+			if(!need_update_list && pthread_mutex_lock(&new_err_jobs_lock) == 0)
 			{
-				pthread_mutex_lock(&new_err_jobs_lock);
 				need_update_list = (new_err_jobs != NULL);
-				pthread_mutex_unlock(&new_err_jobs_lock);
+				(void)pthread_mutex_unlock(&new_err_jobs_lock);
 			}
 			if(need_update_list)
 			{
@@ -494,18 +514,17 @@ free_drained_jobs(bg_job_t **jobs)
 	{
 		bg_job_t *const j = *job;
 
-		if(j->drained)
+		if(j->drained && pthread_spin_lock(&j->status_lock) == 0)
 		{
-			pthread_spin_lock(&j->status_lock);
 			/* If finished, decrement use_count and drop it from the list. */
 			if(!j->running)
 			{
 				--j->use_count;
 				*job = j->err_next;
-				pthread_spin_unlock(&j->status_lock);
+				(void)pthread_spin_unlock(&j->status_lock);
 				continue;
 			}
-			pthread_spin_unlock(&j->status_lock);
+			(void)pthread_spin_unlock(&j->status_lock);
 		}
 
 		job = &j->err_next;
@@ -519,14 +538,20 @@ import_error_jobs(bg_job_t **jobs)
 	bg_job_t *new_jobs;
 
 	/* Add new tasks to internal list, wait if there are no jobs. */
-	pthread_mutex_lock(&new_err_jobs_lock);
+	if(pthread_mutex_lock(&new_err_jobs_lock) != 0)
+	{
+		return;
+	}
 	while(*jobs == NULL && new_err_jobs == NULL)
 	{
-		pthread_cond_wait(&new_err_jobs_cond, &new_err_jobs_lock);
+		if(pthread_cond_wait(&new_err_jobs_cond, &new_err_jobs_lock) != 0)
+		{
+			break;
+		}
 	}
 	new_jobs = new_err_jobs;
 	new_err_jobs = NULL;
-	pthread_mutex_unlock(&new_err_jobs_lock);
+	(void)pthread_mutex_unlock(&new_err_jobs_lock);
 
 	/* Prepend new jobs to the list. */
 	while(new_jobs != NULL)
@@ -617,12 +642,11 @@ report_error_msg(const char title[], const char text[])
 static void
 append_error_msg(bg_job_t *job, const char err_msg[])
 {
-	if(err_msg[0] != '\0')
+	if(err_msg[0] != '\0' && pthread_spin_lock(&job->errors_lock) == 0)
 	{
-		pthread_spin_lock(&job->errors_lock);
 		(void)strappend(&job->errors, &job->errors_len, err_msg);
 		(void)strappend(&job->new_errors, &job->new_errors_len, err_msg);
-		pthread_spin_unlock(&job->errors_lock);
+		(void)pthread_spin_unlock(&job->errors_lock);
 	}
 }
 
@@ -634,7 +658,7 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 	int out_pipe[2];
 	int error_pipe[2];
 
-	if(pipe(out_pipe) != 0)
+	if(out != NULL && pipe(out_pipe) != 0)
 	{
 		show_error_msg("File pipe error", "Error creating pipe");
 		return (pid_t)-1;
@@ -643,8 +667,11 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 	if(pipe(error_pipe) != 0)
 	{
 		show_error_msg("File pipe error", "Error creating pipe");
-		close(out_pipe[0]);
-		close(out_pipe[1]);
+		if(out != NULL)
+		{
+			close(out_pipe[0]);
+			close(out_pipe[1]);
+		}
 		return (pid_t)-1;
 	}
 
@@ -655,8 +682,11 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 
 	if((pid = fork()) == -1)
 	{
-		close(out_pipe[0]);
-		close(out_pipe[1]);
+		if(out != NULL)
+		{
+			close(out_pipe[0]);
+			close(out_pipe[1]);
+		}
 		close(error_pipe[0]);
 		close(error_pipe[1]);
 		return (pid_t)-1;
@@ -664,19 +694,12 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 
 	if(pid == 0)
 	{
-		char *sh;
-		char *sh_flag;
+		if(out != NULL)
+		{
+			bind_pipe_or_die(STDOUT_FILENO, out_pipe[1], out_pipe[0]);
+		}
 
-		close(out_pipe[0]);
-		close(error_pipe[0]);
-		if(dup2(out_pipe[1], STDOUT_FILENO) == -1)
-		{
-			_Exit(EXIT_FAILURE);
-		}
-		if(dup2(error_pipe[1], STDERR_FILENO) == -1)
-		{
-			_Exit(EXIT_FAILURE);
-		}
+		bind_pipe_or_die(STDERR_FILENO, error_pipe[1], error_pipe[0]);
 
 		if(in != NULL)
 		{
@@ -690,25 +713,29 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 			fclose(in);
 		}
 
-		sh = user_sh ? get_execv_path(cfg.shell) : "/bin/sh";
-		sh_flag = user_sh ? cfg.shell_cmd_flag : "-c";
+		char *sh_flag = user_sh ? cfg.shell_cmd_flag : "-c";
 		prepare_for_exec();
-		execvp(sh, make_execv_array(sh, sh_flag, cmd));
+		execvp(get_execv_path(cfg.shell),
+				make_execv_array(cfg.shell, sh_flag, cmd));
 		_Exit(127);
 	}
 
-	close(out_pipe[1]);
+	if(out != NULL)
+	{
+		close(out_pipe[1]);
+		*out = fdopen(out_pipe[0], "r");
+	}
+
 	close(error_pipe[1]);
-	*out = fdopen(out_pipe[0], "r");
 	*err = fdopen(error_pipe[0], "r");
 
 	return pid;
 }
 #else
 /* Runs command in a background and redirects its stdout and stderr streams to
- * file streams which are set.  Input is redirected only if in parameter isn't
- * NULL.  Don't pass pipe for input, it can cause deadlock.  Returns (pid_t)0 or
- * (pid_t)-1 on error. */
+ * file streams which are set.  Input and output are redirected only if the
+ * corresponding parameter isn't NULL.  Don't pass pipe for input, it can cause
+ * deadlock.  Returns (pid_t)0 or (pid_t)-1 on error. */
 static pid_t
 background_and_capture_internal(char cmd[], int user_sh, FILE *in, FILE **out,
 		FILE **err, int out_pipe[2], int err_pipe[2])
@@ -730,7 +757,7 @@ background_and_capture_internal(char cmd[], int user_sh, FILE *in, FILE **out,
 			return (pid_t)-1;
 	}
 
-	if(_dup2(out_pipe[1], _fileno(stdout)) != 0)
+	if(out != NULL && _dup2(out_pipe[1], _fileno(stdout)) != 0)
 		return (pid_t)-1;
 	if(_dup2(err_pipe[1], _fileno(stderr)) != 0)
 		return (pid_t)-1;
@@ -781,11 +808,17 @@ background_and_capture_internal(char cmd[], int user_sh, FILE *in, FILE **out,
 		return (pid_t)-1;
 	}
 
-	if((*out = _fdopen(out_pipe[0], "r")) == NULL)
+	if(out != NULL && (*out = _fdopen(out_pipe[0], "r")) == NULL)
+	{
 		return (pid_t)-1;
+	}
+
 	if((*err = _fdopen(err_pipe[0], "r")) == NULL)
 	{
-		fclose(*out);
+		if(out != NULL)
+		{
+			fclose(*out);
+		}
 		return (pid_t)-1;
 	}
 
@@ -800,7 +833,7 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 	int err_fd, err_pipe[2];
 	pid_t pid;
 
-	if(_pipe(out_pipe, 512, O_NOINHERIT) != 0)
+	if(out != NULL && _pipe(out_pipe, 512, O_NOINHERIT) != 0)
 	{
 		show_error_msg("File pipe error", "Error creating pipe");
 		return (pid_t)-1;
@@ -809,8 +842,11 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 	if(_pipe(err_pipe, 512, O_NOINHERIT) != 0)
 	{
 		show_error_msg("File pipe error", "Error creating pipe");
-		close(out_pipe[0]);
-		close(out_pipe[1]);
+		if(out != NULL)
+		{
+			close(out_pipe[0]);
+			close(out_pipe[1]);
+		}
 		return (pid_t)-1;
 	}
 
@@ -821,7 +857,10 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 	pid = background_and_capture_internal(cmd, user_sh, in, out, err, out_pipe,
 			err_pipe);
 
-	_close(out_pipe[1]);
+	if(out != NULL)
+	{
+		_close(out_pipe[1]);
+	}
 	_close(err_pipe[1]);
 
 	_dup2(in_fd, _fileno(stdin));
@@ -830,7 +869,10 @@ bg_run_and_capture(char cmd[], int user_sh, FILE *in, FILE **out, FILE **err)
 
 	if(pid == (pid_t)-1)
 	{
-		_close(out_pipe[0]);
+		if(out != NULL)
+		{
+			_close(out_pipe[0]);
+		}
 		_close(err_pipe[0]);
 	}
 
@@ -908,7 +950,7 @@ launch_external(const char cmd[], BgJobFlags flags, ShellRequester by)
 	const int merge_streams = (capture_output && (flags & BJF_MERGE_STREAMS));
 
 #ifndef _WIN32
-	const int keep_session = (flags & BJF_KEEP_SESSION);
+	const int keep_in_fg = (flags & BJF_KEEP_IN_FG);
 
 	pid_t pid;
 	int input_pipe[2] = { -1, -1 };
@@ -977,32 +1019,22 @@ launch_external(const char cmd[], BgJobFlags flags, ShellRequester by)
 		}
 		close(STDIN_FILENO);
 		close(STDOUT_FILENO);
-		/* Close read end of pipe. */
+
+		/* Close original error pipe descriptors. */
 		if(error_pipe[0] != -1)
 		{
 			close(error_pipe[0]);
+			close(error_pipe[1]);
 		}
 
 		if(supply_input)
 		{
-			if(dup2(input_pipe[0], STDIN_FILENO) == -1)
-			{
-				perror("dup2");
-				_Exit(EXIT_FAILURE);
-			}
-			/* Close write end of pipe. */
-			close(input_pipe[1]);
+			bind_pipe_or_die(STDIN_FILENO, input_pipe[0], input_pipe[1]);
 		}
 
 		if(capture_output)
 		{
-			if(dup2(output_pipe[1], STDOUT_FILENO) == -1)
-			{
-				perror("dup2");
-				_Exit(EXIT_FAILURE);
-			}
-			/* Close read end of pipe. */
-			close(output_pipe[0]);
+			bind_pipe_or_die(STDOUT_FILENO, output_pipe[1], output_pipe[0]);
 		}
 
 		/* Attach stdin and optionally stdout to /dev/null. */
@@ -1019,19 +1051,15 @@ launch_external(const char cmd[], BgJobFlags flags, ShellRequester by)
 				perror("dup2 for stdout");
 				_Exit(EXIT_FAILURE);
 			}
-		}
-
-		if(keep_session)
-		{
-			if(setpgid(0, 0) == -1)
+			if(nullfd != STDIN_FILENO && nullfd != STDOUT_FILENO)
 			{
-				perror("setpgid");
-				_Exit(EXIT_FAILURE);
+				close(nullfd);
 			}
 		}
+
 		/* setsid() creates process group as well and doesn't work if current
 		 * process is a group leader, so don't do setpgid(). */
-		else if(setsid() == (pid_t)-1)
+		if(!keep_in_fg && setsid() == (pid_t)-1)
 		{
 			perror("setsid");
 			_Exit(EXIT_FAILURE);
@@ -1250,10 +1278,12 @@ bg_execute(const char descr[], const char op_descr[], int total, int important,
 	if(pthread_create(&id, NULL, &background_task_bootstrap, task_args) != 0)
 	{
 		/* Mark job as finished with error. */
-		pthread_spin_lock(&task_args->job->status_lock);
-		task_args->job->running = 0;
-		task_args->job->exit_code = 1;
-		pthread_spin_unlock(&task_args->job->status_lock);
+		if(pthread_spin_lock(&task_args->job->status_lock) == 0)
+		{
+			task_args->job->running = 0;
+			task_args->job->exit_code = 1;
+			(void)pthread_spin_unlock(&task_args->job->status_lock);
+		}
 
 		free(task_args);
 		ret = 1;
@@ -1291,12 +1321,15 @@ add_background_job(pid_t pid, const char cmd[], uintptr_t err, uintptr_t data,
 	bg_job_t *new = malloc(sizeof(*new));
 	if(new == NULL)
 	{
-		show_error_msg("Memory error", "Unable to allocate enough memory");
 		return NULL;
 	}
+	new->cmd = strdup(cmd);
+	if(new->cmd == NULL)
+	{
+		goto free_new;
+	}
 	new->type = type;
 	new->pid = pid;
-	new->cmd = strdup(cmd);
 	new->next = bg_jobs;
 	new->skip_errors = 0;
 	new->new_errors = NULL;
@@ -1305,8 +1338,20 @@ add_background_job(pid_t pid, const char cmd[], uintptr_t err, uintptr_t data,
 	new->errors_len = 0U;
 	new->cancelled = 0;
 
-	pthread_spin_init(&new->errors_lock, PTHREAD_PROCESS_PRIVATE);
-	pthread_spin_init(&new->status_lock, PTHREAD_PROCESS_PRIVATE);
+	if(pthread_spin_init(&new->errors_lock, PTHREAD_PROCESS_PRIVATE) != 0)
+	{
+		goto free_cmd;
+	}
+	if(pthread_spin_init(&new->status_lock, PTHREAD_PROCESS_PRIVATE) != 0)
+	{
+		goto free_errors_lock;
+	}
+	if(with_bg_op &&
+			pthread_spin_init(&new->bg_op_lock, PTHREAD_PROCESS_PRIVATE) != 0)
+	{
+		goto free_status_lock;
+	}
+
 	new->running = 1;
 	new->use_count = 0;
 	new->exit_code = -1;
@@ -1314,6 +1359,9 @@ add_background_job(pid_t pid, const char cmd[], uintptr_t err, uintptr_t data,
 	new->input = NULL;
 	new->output = NULL;
 
+	new->exit_cb = NULL;
+	new->exit_cb_arg = NULL;
+
 #ifndef _WIN32
 	new->err_stream = (int)err;
 #else
@@ -1325,19 +1373,19 @@ add_background_job(pid_t pid, const char cmd[], uintptr_t err, uintptr_t data,
 	if(new->err_stream != NO_JOB_ID)
 	{
 		++new->use_count;
-		pthread_mutex_lock(&new_err_jobs_lock);
+
+		if(pthread_mutex_lock(&new_err_jobs_lock) != 0)
+		{
+			goto free_bg_op_lock;
+		}
 		new->err_next = new_err_jobs;
 		new_err_jobs = new;
-		pthread_mutex_unlock(&new_err_jobs_lock);
-		pthread_cond_signal(&new_err_jobs_cond);
+		(void)pthread_mutex_unlock(&new_err_jobs_lock);
+		(void)pthread_cond_signal(&new_err_jobs_cond);
 	}
 
 	new->with_bg_op = with_bg_op;
 	new->on_job_bar = 0;
-	if(new->with_bg_op)
-	{
-		pthread_spin_init(&new->bg_op_lock, PTHREAD_PROCESS_PRIVATE);
-	}
 	new->bg_op.total = 0;
 	new->bg_op.done = 0;
 	new->bg_op.progress = -1;
@@ -1348,6 +1396,21 @@ add_background_job(pid_t pid, const char cmd[], uintptr_t err, uintptr_t data,
 
 	bg_jobs = new;
 	return new;
+
+free_bg_op_lock:
+	if(with_bg_op)
+	{
+		(void)pthread_spin_destroy(&new->bg_op_lock);
+	}
+free_status_lock:
+	(void)pthread_spin_destroy(&new->status_lock);
+free_errors_lock:
+	(void)pthread_spin_destroy(&new->errors_lock);
+free_cmd:
+	free(new->cmd);
+free_new:
+	free(new);
+	return NULL;
 }
 
 /* pthreads entry point for a new background task.  Performs correct
@@ -1360,35 +1423,22 @@ background_task_bootstrap(void *arg)
 
 	(void)pthread_detach(pthread_self());
 	block_all_thread_signals();
-	set_current_job(task_args->job);
 
-	task_args->func(&task_args->job->bg_op, task_args->args);
-
-	/* Mark task as finished normally. */
-	mark_job_finished(task_args->job, 0);
+	if(pthread_setspecific(current_job, task_args->job) == 0)
+	{
+		task_args->func(&task_args->job->bg_op, task_args->args);
+		mark_job_finished(task_args->job, /*exit_code=*/0);
+	}
+	else
+	{
+		mark_job_finished(task_args->job, /*exit_code=*/1);
+	}
 
 	free(task_args);
 
 	return NULL;
 }
 
-/* Stores pointer to the job in a thread-local storage. */
-static void
-set_current_job(bg_job_t *job)
-{
-	static pthread_once_t once = PTHREAD_ONCE_INIT;
-	pthread_once(&once, &make_current_job_key);
-
-	(void)pthread_setspecific(current_job, job);
-}
-
-/* current_job initializer for pthread_once(). */
-static void
-make_current_job_key(void)
-{
-	(void)pthread_key_create(&current_job, NULL);
-}
-
 int
 bg_has_active_jobs(int important_only)
 {
@@ -1406,6 +1456,13 @@ bg_has_active_jobs(int important_only)
 	return running;
 }
 
+void
+bg_job_set_exit_cb(bg_job_t *job, bg_job_exit_func cb, void *arg)
+{
+	job->exit_cb = cb;
+	job->exit_cb_arg = arg;
+}
+
 int
 bg_job_cancel(bg_job_t *job)
 {
@@ -1478,19 +1535,25 @@ bg_job_terminate(bg_job_t *job)
 int
 bg_job_is_running(bg_job_t *job)
 {
-	pthread_spin_lock(&job->status_lock);
+	if(pthread_spin_lock(&job->status_lock) != 0)
+	{
+		return 1;
+	}
 	int running = job->running;
-	pthread_spin_unlock(&job->status_lock);
+	(void)pthread_spin_unlock(&job->status_lock);
 	return (running && update_job_status(job));
 }
 
 int
 bg_job_was_killed(bg_job_t *job)
 {
-	pthread_spin_lock(&job->status_lock);
+	if(pthread_spin_lock(&job->status_lock) != 0)
+	{
+		return 0;
+	}
 	int running = job->running;
 	int exit_code = job->exit_code;
-	pthread_spin_unlock(&job->status_lock);
+	(void)pthread_spin_unlock(&job->status_lock);
 	return (!running && exit_code >= 0);
 }
 
@@ -1520,7 +1583,7 @@ bg_job_wait(bg_job_t *job)
 	}
 
 #ifndef _WIN32
-	int status = get_proc_exit_status(job->pid);
+	int status = get_proc_exit_status(job->pid, &no_cancellation);
 	if(status == -1)
 	{
 		return 1;
@@ -1567,34 +1630,42 @@ update_job_status(bg_job_t *job)
 static void
 mark_job_finished(bg_job_t *job, int exit_code)
 {
-	pthread_spin_lock(&job->status_lock);
-	job->running = 0;
-	job->exit_code = exit_code;
-	pthread_spin_unlock(&job->status_lock);
+	if(pthread_spin_lock(&job->status_lock) == 0)
+	{
+		job->running = 0;
+		job->exit_code = exit_code;
+		(void)pthread_spin_unlock(&job->status_lock);
+	}
 }
 
 void
 bg_job_incref(bg_job_t *job)
 {
-	pthread_spin_lock(&job->status_lock);
-	++job->use_count;
-	pthread_spin_unlock(&job->status_lock);
+	if(pthread_spin_lock(&job->status_lock) == 0)
+	{
+		++job->use_count;
+		(void)pthread_spin_unlock(&job->status_lock);
+	}
 }
 
 void
 bg_job_decref(bg_job_t *job)
 {
-	pthread_spin_lock(&job->status_lock);
-	--job->use_count;
-	pthread_spin_unlock(&job->status_lock);
+	if(pthread_spin_lock(&job->status_lock) == 0)
+	{
+		--job->use_count;
+		assert(job->use_count >= 0 && "Excessive bg_job_decref() call!");
+		(void)pthread_spin_unlock(&job->status_lock);
+	}
 }
 
-void
+int
 bg_op_lock(bg_op_t *bg_op)
 {
 	bg_job_t *const job = STRUCT_FROM_FIELD(bg_job_t, bg_op, bg_op);
 	assert(job->with_bg_op && "This function requires bg_op data.");
-	pthread_spin_lock(&job->bg_op_lock);
+
+	return (pthread_spin_lock(&job->bg_op_lock) == 0);
 }
 
 void
@@ -1602,7 +1673,10 @@ bg_op_unlock(bg_op_t *bg_op)
 {
 	bg_job_t *const job = STRUCT_FROM_FIELD(bg_job_t, bg_op, bg_op);
 	assert(job->with_bg_op && "This function requires bg_op data.");
-	pthread_spin_unlock(&job->bg_op_lock);
+
+	int error = pthread_spin_unlock(&job->bg_op_lock);
+	assert(error == 0 && "Unlock failure in bg_op_lock()");
+	(void)error;
 }
 
 void
@@ -1614,11 +1688,13 @@ bg_op_changed(bg_op_t *bg_op)
 void
 bg_op_set_descr(bg_op_t *bg_op, const char descr[])
 {
-	bg_op_lock(bg_op);
-	replace_string(&bg_op->descr, descr);
-	bg_op_unlock(bg_op);
+	if(bg_op_lock(bg_op))
+	{
+		replace_string(&bg_op->descr, descr);
+		bg_op_unlock(bg_op);
 
-	bg_op_changed(bg_op);
+		bg_op_changed(bg_op);
+	}
 }
 
 /* Convenience method to cancel background job.  Returns previous version of the
@@ -1626,26 +1702,27 @@ bg_op_set_descr(bg_op_t *bg_op, const char descr[])
 static int
 bg_op_cancel(bg_op_t *bg_op)
 {
-	int was_cancelled;
+	int was_cancelled = 0;
+	if(bg_op_lock(bg_op))
+	{
+		was_cancelled = bg_op->cancelled;
+		bg_op->cancelled = 1;
+		bg_op_unlock(bg_op);
 
-	bg_op_lock(bg_op);
-	was_cancelled = bg_op->cancelled;
-	bg_op->cancelled = 1;
-	bg_op_unlock(bg_op);
-
-	bg_op_changed(bg_op);
+		bg_op_changed(bg_op);
+	}
 	return was_cancelled;
 }
 
 int
 bg_op_cancelled(bg_op_t *bg_op)
 {
-	int cancelled;
-
-	bg_op_lock(bg_op);
-	cancelled = bg_op->cancelled;
-	bg_op_unlock(bg_op);
-
+	int cancelled = 0;
+	if(bg_op_lock(bg_op))
+	{
+		cancelled = bg_op->cancelled;
+		bg_op_unlock(bg_op);
+	}
 	return cancelled;
 }
 
