@@ -35,6 +35,7 @@
 #include "engine/completion.h"
 #include "engine/keys.h"
 #include "engine/mode.h"
+#include "lua/vlua.h"
 #include "modes/dialogs/msg_dialog.h"
 #include "modes/modes.h"
 #include "modes/wk.h"
@@ -150,6 +151,15 @@ event_loop(const int *quit, int manage_marking)
 
 			bg_check();
 
+			/* Lua might not be initialized in tests. */
+			if(input_buf_pos == 0 && !wait_for_enter && vle_mode_is(NORMAL_MODE) &&
+					curr_stats.vlua != NULL)
+			{
+				/* We're not waiting for anything, so side-effects of callbacks
+				 * shouldn't be disruptive to the user. */
+				vlua_process_callbacks(curr_stats.vlua);
+			}
+
 			got_input = (get_char_async_loop(status_bar, &c, actual_timeout) != ERR);
 
 			/* If suggestion delay timed out, reset it and wait the rest of the
@@ -214,7 +224,7 @@ event_loop(const int *quit, int manage_marking)
 			{
 				/* Recover from input buffer overflow by resetting its contents. */
 				reset_input_buf(input_buf, &input_buf_pos);
-				clear_input_bar();
+				modes_input_bar_clear();
 				continue;
 			}
 		}
@@ -272,7 +282,7 @@ event_loop(const int *quit, int manage_marking)
 
 				if(got_input)
 				{
-					modupd_input_bar(input_buf);
+					modes_input_bar_update(input_buf);
 				}
 
 				if(last_result == KEYS_WAIT_SHORT && wcscmp(input_buf, L"\033") == 0)
@@ -282,14 +292,10 @@ event_loop(const int *quit, int manage_marking)
 
 				if(counter > 0)
 				{
-					clear_input_bar();
+					modes_input_bar_clear();
 				}
 
-				if(!curr_stats.save_msg && curr_view->selected_files &&
-						!vle_mode_is(CMDLINE_MODE))
-				{
-					print_selected_msg();
-				}
+				modes_statusbar_update();
 				continue;
 			}
 		}
@@ -299,7 +305,7 @@ event_loop(const int *quit, int manage_marking)
 		process_scheduled_updates();
 
 		reset_input_buf(input_buf, &input_buf_pos);
-		clear_input_bar();
+		modes_input_bar_clear();
 
 		if(ui_sb_multiline())
 		{
@@ -566,9 +572,9 @@ static int
 should_check_views_for_changes(void)
 {
 	return !ui_sb_multiline()
-	    && !is_in_menu_like_mode()
+	    && !modes_is_menu_like()
 	    && !modes_is_dialog_like()
-	    && !vle_mode_is(CMDLINE_MODE)
+	    && !modes_is_cmdline_like()
 	    && !suggestions_are_visible;
 }
 
