@@ -108,7 +108,9 @@ static void cmd_zh(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_zl(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_zt(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_zz(key_info_t key_info, keys_info_t *keys_info);
+static void handle_mouse_event(key_info_t key_info, keys_info_t *keys_info);
 static int all_lines_visible(const menu_data_t *menu);
+
 static int goto_cmd(const cmd_info_t *cmd_info);
 static int nohlsearch_cmd(const cmd_info_t *cmd_info);
 static int quit_cmd(const cmd_info_t *cmd_info);
@@ -174,6 +176,7 @@ static keys_add_info_t builtin_cmds[] = {
 	{{WC_z, K(KEY_LEFT)},  {{&cmd_zh},     .descr = "scroll one column left"}},
 	{{WC_z, K(KEY_RIGHT)}, {{&cmd_zl},     .descr = "scroll one column right"}},
 #endif /* ENABLE_EXTENDED_KEYS */
+	{{K(KEY_MOUSE)}, {{&handle_mouse_event}, FOLLOWED_BY_NONE}},
 };
 
 /* Specification of builtin commands. */
@@ -564,7 +567,7 @@ cmd_slash(key_info_t key_info, keys_info_t *keys_info)
 	last_search_backward = 0;
 	menus_search_reset(menu->state, last_search_backward,
 			def_count(key_info.count));
-	modcline_enter(CLS_MENU_FSEARCH, "", menu);
+	modcline_in_menu(CLS_MENU_FSEARCH, menu);
 }
 
 /* Jump to percent of list. */
@@ -587,7 +590,7 @@ cmd_colon(key_info_t key_info, keys_info_t *keys_info)
 	cmds_conf.begin = 1;
 	cmds_conf.current = menu->pos;
 	cmds_conf.end = menu->len;
-	modcline_enter(CLS_MENU_COMMAND, "", menu);
+	modcline_in_menu(CLS_MENU_COMMAND, menu);
 }
 
 static void
@@ -596,7 +599,7 @@ cmd_qmark(key_info_t key_info, keys_info_t *keys_info)
 	last_search_backward = 1;
 	menus_search_reset(menu->state, last_search_backward,
 			def_count(key_info.count));
-	modcline_enter(CLS_MENU_BSEARCH, "", menu);
+	modcline_in_menu(CLS_MENU_BSEARCH, menu);
 }
 
 /* Populates very custom (unsorted) view with list of files. */
@@ -844,8 +847,10 @@ cmd_v(key_info_t key_info, keys_info_t *keys_info)
 
 	if(!qf)
 	{
-		char *const arg = shell_like_escape("+exe 'bd!|args' "
-				"join(map(getline('1','$'),'fnameescape(v:val)'))", 0);
+		ShellType shell_type = (get_env_type() == ET_UNIX ? ST_POSIX : ST_CMD);
+		char *const arg = shell_arg_escape("+exe 'bd!|args' "
+				"join(map(getline('1','$'),'fnameescape(v:val)'))",
+				shell_type);
 		cmd = format_str("%s %s +argument%d -", vi_cmd, arg, menu->pos + 1);
 		free(arg);
 	}
@@ -1051,7 +1056,7 @@ modmenu_morph_into_cline(CmdLineSubmode submode, const char input[],
 	}
 
 	leave_menu_mode(0);
-	modcline_enter(submode, input_copy, NULL);
+	modcline_enter(submode, input_copy);
 
 	free(input_copy);
 }
@@ -1083,7 +1088,7 @@ leave_menu_mode(int reset_selection)
 void
 modmenu_run_command(const char cmd[])
 {
-	if(exec_command(cmd, view, CIT_COMMAND) < 0)
+	if(cmds_dispatch1(cmd, view, CIT_COMMAND) < 0)
 	{
 		ui_sb_err("An error occurred while trying to execute command");
 	}
@@ -1097,5 +1102,44 @@ menu_get_current(void)
 	return menu;
 }
 
+/* Processes events from the mouse. */
+static void
+handle_mouse_event(key_info_t key_info, keys_info_t *keys_info)
+{
+	MEVENT e;
+	if(ui_get_mouse(&e) != OK)
+	{
+		return;
+	}
+
+	if(!wenclose(menu_win, e.y, e.x))
+	{
+		return;
+	}
+
+	if(e.bstate & BUTTON1_PRESSED)
+	{
+		wmouse_trafo(menu_win, &e.y, &e.x, FALSE);
+
+		int old_pos = menu->pos;
+
+		menus_erase_current(menu->state);
+		menus_set_pos(menu->state, menu->top + e.y - 1);
+		ui_refresh_win(menu_win);
+
+		if(menu->pos == old_pos)
+		{
+			cmd_return(key_info, keys_info);
+		}
+	}
+	else if(e.bstate & BUTTON4_PRESSED)
+	{
+		cmd_ctrl_y(key_info, keys_info);
+	}
+	else if(e.bstate & (BUTTON2_PRESSED | BUTTON5_PRESSED))
+	{
+		cmd_ctrl_e(key_info, keys_info);
+	}
+}
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
 /* vim: set cinoptions+=t0 filetype=c : */
