@@ -40,6 +40,7 @@
 #include "io/ioeta.h"
 #include "io/iop.h"
 #include "io/ior.h"
+#include "lua/vlua.h"
 #include "modes/dialogs/msg_dialog.h"
 #include "ui/cancellation.h"
 #include "utils/cancellation.h"
@@ -130,6 +131,7 @@ static OpsResult op_rmdir(ops_t *ops, void *data, const char src[],
 static OpsResult op_mkfile(ops_t *ops, void *data, const char src[],
 		const char dst[]);
 static int ops_uses_syscalls(const ops_t *ops);
+static ShellType ops_shell_type(const ops_t *ops);
 static OpsResult exec_io_op(ops_t *ops, IoRes (*func)(io_args_t *),
 		io_args_t *args, int cancellable);
 static int confirm_overwrite(io_args_t *args, const char src[],
@@ -142,6 +144,7 @@ static int ui_cancellation_hook(void *arg);
 #ifndef _WIN32
 static OpsResult run_operation_command(ops_t *ops, char cmd[], int cancellable);
 #endif
+static int ops_runs_in_bg(const ops_t *ops);
 static int bg_cancellation_hook(void *arg);
 static OpsResult result_from_code(int exit_code);
 
@@ -296,7 +299,28 @@ OpsResult
 perform_operation(OPS op, ops_t *ops, void *data, const char src[],
 		const char dst[])
 {
-	return op_funcs[op](ops, data, src, dst);
+	if(ops_runs_in_bg(ops))
+	{
+		/* Not reporting events from background jobs. */
+		return op_funcs[op](ops, data, src, dst);
+	}
+
+	int dir = 0;
+	if(ONE_OF(op, OP_MKDIR, OP_RMDIR))
+	{
+		dir = 1;
+	}
+	else if(!ONE_OF(op, OP_NONE, OP_USR, OP_MKFILE, OP_SYMLINK, OP_SYMLINK2))
+	{
+		dir = is_dir(src);
+	}
+
+	OpsResult status = op_funcs[op](ops, data, src, dst);
+	if(status == OPS_SUCCEEDED)
+	{
+		vlua_events_app_fsop(curr_stats.vlua, op, src, dst, data, dir);
+	}
+	return status;
 }
 
 static OpsResult
@@ -340,7 +364,7 @@ op_removesl(ops_t *ops, void *data, const char src[], const char dst[])
 		char cmd[2*PATH_MAX + 1];
 		const int cancellable = (data == NULL);
 
-		escaped = shell_like_escape(src, 0);
+		escaped = shell_arg_escape(src, ops_shell_type(ops));
 		if(escaped == NULL)
 		{
 			return OPS_FAILED;
@@ -352,9 +376,15 @@ op_removesl(ops_t *ops, void *data, const char src[], const char dst[])
 		LOG_INFO_MSG("Running trash command: \"%s\"", cmd);
 		return run_operation_command(ops, cmd, cancellable);
 #else
+		char *src_copy = strdup(src);
+		internal_to_system_slashes(src_copy);
+
+		char *escaped_src = shell_arg_escape(src_copy, ST_CMD);
+		free(src_copy);
+
 		char cmd[PATH_MAX*2 + 1];
-		snprintf(cmd, sizeof(cmd), "%s \"%s\"", delete_prg, src);
-		internal_to_system_slashes(cmd);
+		snprintf(cmd, sizeof(cmd), "%s %s", delete_prg, escaped_src);
+		free(escaped_src);
 
 		return result_from_code(os_system(cmd));
 #endif
@@ -367,7 +397,7 @@ op_removesl(ops_t *ops, void *data, const char src[], const char dst[])
 		char cmd[16 + PATH_MAX];
 		const int cancellable = data == NULL;
 
-		escaped = shell_like_escape(src, 0);
+		escaped = shell_arg_escape(src, ops_shell_type(ops));
 		if(escaped == NULL)
 			return OPS_FAILED;
 
@@ -484,8 +514,8 @@ op_cp(ops_t *ops, void *data, const char src[], const char dst[],
 		char cmd[6 + PATH_MAX*2 + 1];
 		const int cancellable = (data == NULL);
 
-		escaped_src = shell_like_escape(src, 0);
-		escaped_dst = shell_like_escape(dst, 0);
+		escaped_src = shell_arg_escape(src, ops_shell_type(ops));
+		escaped_dst = shell_arg_escape(dst, ops_shell_type(ops));
 		if(escaped_src == NULL || escaped_dst == NULL)
 		{
 			free(escaped_dst);
@@ -509,9 +539,22 @@ op_cp(ops_t *ops, void *data, const char src[], const char dst[],
 
 		if(is_dir(src))
 		{
+			char *src_copy = strdup(src);
+			internal_to_system_slashes(src_copy);
+
+			char *dst_copy = strdup(dst);
+			internal_to_system_slashes(dst_copy);
+
+			char *escaped_src = shell_arg_escape(src_copy, ST_CMD);
+			free(src_copy);
+
+			char *escaped_dst = shell_arg_escape(dst_copy, ST_CMD);
+			free(dst_copy);
+
 			char cmd[6 + PATH_MAX*2 + 1];
-			snprintf(cmd, sizeof(cmd), "xcopy \"%s\" \"%s\" ", src, dst);
-			internal_to_system_slashes(cmd);
+			snprintf(cmd, sizeof(cmd), "xcopy %s %s ", escaped_src, escaped_dst);
+			free(escaped_src);
+			free(escaped_dst);
 
 			if(is_vista_and_above())
 				strcat(cmd, "/B ");
@@ -598,8 +641,8 @@ op_mv(ops_t *ops, void *data, const char src[], const char dst[],
 			return OPS_FAILED;
 		}
 
-		escaped_src = shell_like_escape(src, 0);
-		escaped_dst = shell_like_escape(dst, 0);
+		escaped_src = shell_arg_escape(src, ops_shell_type(ops));
+		escaped_dst = shell_arg_escape(dst, ops_shell_type(ops));
 		if(escaped_src == NULL || escaped_dst == NULL)
 		{
 			free(escaped_dst);
@@ -651,7 +694,9 @@ op_mv(ops_t *ops, void *data, const char src[], const char dst[],
 		result = exec_io_op(ops, &ior_mv, &args, data == NULL);
 	}
 
-	if(result == OPS_SUCCEEDED)
+	/* Accounting for background jobs might take some kind of a queue of events
+	 * to handle both internal updates and app.fsop Lua event. */
+	if(result == OPS_SUCCEEDED && !ops_runs_in_bg(ops))
 	{
 		trash_file_moved(src, dst);
 		bmarks_file_moved(src, dst);
@@ -683,7 +728,7 @@ op_chown(ops_t *ops, void *data, const char src[], const char dst[])
 	char *escaped;
 	uid_t uid = (uid_t)(long)data;
 
-	escaped = shell_like_escape(src, 0);
+	escaped = shell_arg_escape(src, ops_shell_type(ops));
 	snprintf(cmd, sizeof(cmd), "chown -fR %u %s", uid, escaped);
 	free(escaped);
 
@@ -702,7 +747,7 @@ op_chgrp(ops_t *ops, void *data, const char src[], const char dst[])
 	char *escaped;
 	gid_t gid = (gid_t)(long)data;
 
-	escaped = shell_like_escape(src, 0);
+	escaped = shell_arg_escape(src, ops_shell_type(ops));
 	snprintf(cmd, sizeof(cmd), "chown -fR :%u %s", gid, escaped);
 	free(escaped);
 
@@ -720,7 +765,7 @@ op_chmod(ops_t *ops, void *data, const char src[], const char dst[])
 	char cmd[128 + PATH_MAX];
 	char *escaped;
 
-	escaped = shell_like_escape(src, 0);
+	escaped = shell_arg_escape(src, ops_shell_type(ops));
 	snprintf(cmd, sizeof(cmd), "chmod %s %s", (char *)data, escaped);
 	free(escaped);
 
@@ -734,7 +779,7 @@ op_chmodr(ops_t *ops, void *data, const char src[], const char dst[])
 	char cmd[128 + PATH_MAX];
 	char *escaped;
 
-	escaped = shell_like_escape(src, 0);
+	escaped = shell_arg_escape(src, ops_shell_type(ops));
 	snprintf(cmd, sizeof(cmd), "chmod -R %s %s", (char *)data, escaped);
 	free(escaped);
 
@@ -792,17 +837,29 @@ op_symlink(ops_t *ops, void *data, const char src[], const char dst[])
 {
 	if(!ops_uses_syscalls(ops))
 	{
-		char *escaped_src, *escaped_dst;
 		char cmd[6 + PATH_MAX*2 + 1];
 		OpsResult result;
 
-#ifndef _WIN32
-		escaped_src = shell_like_escape(src, 0);
-		escaped_dst = shell_like_escape(dst, 0);
-#else
-		escaped_src = strdup(enclose_in_dquotes(src, ops->shell_type));
-		escaped_dst = strdup(enclose_in_dquotes(dst, ops->shell_type));
-#endif
+		char *src_copy = strdup(src);
+		char *dst_copy = strdup(dst);
+		if(src_copy == NULL || dst_copy == NULL)
+		{
+			free(src_copy);
+			free(dst_copy);
+			return OPS_FAILED;
+		}
+
+		internal_to_system_slashes(src_copy);
+		internal_to_system_slashes(dst_copy);
+
+		ShellType shell_type = (get_env_type() == ET_UNIX) ? ops_shell_type(ops)
+		                                                   : ST_CMD;
+
+		char *escaped_src = shell_arg_escape(src_copy, shell_type);
+		free(src_copy);
+
+		char *escaped_dst = shell_arg_escape(dst_copy, shell_type);
+		free(dst_copy);
 
 		if(escaped_src == NULL || escaped_dst == NULL)
 		{
@@ -824,8 +881,17 @@ op_symlink(ops_t *ops, void *data, const char src[], const char dst[])
 			return OPS_FAILED;
 		}
 
-		snprintf(cmd, sizeof(cmd), "%s\\win_helper -s %s %s", exe_dir, escaped_src,
+		internal_to_system_slashes(exe_dir);
+
+		char helper[PATH_MAX + 2];
+		snprintf(helper, sizeof(helper), "%s\\win_helper", exe_dir);
+
+		char *escaped_helper = shell_arg_escape(helper, shell_type);
+
+		snprintf(cmd, sizeof(cmd), "%s -s %s %s", escaped_helper, escaped_src,
 				escaped_dst);
+		free(escaped_helper);
+
 		result = result_from_code(os_system(cmd));
 #endif
 
@@ -851,7 +917,7 @@ op_mkdir(ops_t *ops, void *data, const char src[], const char dst[])
 		char cmd[128 + PATH_MAX];
 		char *escaped;
 
-		escaped = shell_like_escape(src, 0);
+		escaped = shell_arg_escape(src, ops_shell_type(ops));
 		snprintf(cmd, sizeof(cmd), "mkdir %s %s", (data == NULL) ? "" : "-p",
 				escaped);
 		free(escaped);
@@ -908,7 +974,7 @@ op_rmdir(ops_t *ops, void *data, const char src[], const char dst[])
 		char cmd[128 + PATH_MAX];
 		char *escaped;
 
-		escaped = shell_like_escape(src, 0);
+		escaped = shell_arg_escape(src, ops_shell_type(ops));
 		snprintf(cmd, sizeof(cmd), "rmdir %s", escaped);
 		free(escaped);
 		LOG_INFO_MSG("Running rmdir command: \"%s\"", cmd);
@@ -936,7 +1002,7 @@ op_mkfile(ops_t *ops, void *data, const char src[], const char dst[])
 		char cmd[128 + PATH_MAX];
 		char *escaped;
 
-		escaped = shell_like_escape(src, 0);
+		escaped = shell_arg_escape(src, ops_shell_type(ops));
 		snprintf(cmd, sizeof(cmd), "touch %s", escaped);
 		free(escaped);
 		LOG_INFO_MSG("Running touch command: \"%s\"", cmd);
@@ -972,6 +1038,13 @@ ops_uses_syscalls(const ops_t *ops)
 	return ops == NULL ? cfg.use_system_calls : ops->use_system_calls;
 }
 
+/* Retrieves shell type for the operation.  Returns the type. */
+static ShellType
+ops_shell_type(const ops_t *ops)
+{
+	return (ops == NULL ? curr_stats.shell_type : (ShellType)ops->shell_type);
+}
+
 /* Executes i/o operation with some predefined pre/post actions.  Returns
  * status. */
 static OpsResult
@@ -993,7 +1066,7 @@ exec_io_op(ops_t *ops, IoRes (*func)(io_args_t *), io_args_t *args,
 
 	if(cancellable)
 	{
-		if(ops != NULL && ops->bg)
+		if(ops_runs_in_bg(ops))
 		{
 			args->cancellation.arg = ops->bg_op;
 			args->cancellation.hook = &bg_cancellation_hook;
@@ -1009,11 +1082,13 @@ exec_io_op(ops_t *ops, IoRes (*func)(io_args_t *), io_args_t *args,
 
 	curr_ops = ops;
 	OpsResult result = OPS_FAILED;
-	switch(func(args))
+	IoRes io_res = func(args);
+	switch(io_res)
 	{
 		case IO_RES_SUCCEEDED: result = OPS_SUCCEEDED; break;
 		case IO_RES_SKIPPED:   result = OPS_SKIPPED; break;
 		case IO_RES_FAILED:    result = OPS_FAILED; break;
+		case IO_RES_ABORTED:   result = OPS_FAILED; break;
 	}
 	curr_ops = NULL;
 
@@ -1024,6 +1099,11 @@ exec_io_op(ops_t *ops, IoRes (*func)(io_args_t *), io_args_t *args,
 
 	if(ops != NULL)
 	{
+		if(io_res == IO_RES_ABORTED)
+		{
+			ops->aborted = 1;
+		}
+
 		size_t len = (ops->errors == NULL) ? 0U : strlen(ops->errors);
 		char *const suffix = ioe_errlst_to_str(&args->result.errors);
 
@@ -1047,10 +1127,10 @@ confirm_overwrite(io_args_t *args, const char src[], const char dst[])
 {
 	/* TODO: think about adding "append" and "rename" options here. */
 	static const response_variant responses[] = {
-		{ .key = 'y', .descr = "[y]es", },
-		{ .key = 'Y', .descr = "[Y]es for all", },
-		{ .key = 'n', .descr = "[n]o", },
-		{ .key = 'N', .descr = "[N]o for all", },
+		{ .key = 'y', .descr = "[y]es/", },
+		{ .key = 'Y', .descr = "[Y]es for all/", },
+		{ .key = 'n', .descr = "[n]o/", },
+		{ .key = 'N', .descr = "[N]o for all/", },
 		{ },
 	};
 
@@ -1115,9 +1195,9 @@ static IoErrCbResult
 dispatch_error(io_args_t *args, const ioe_err_t *err)
 {
 	static const response_variant responses[] = {
-		{ .key = 'r', .descr = "[r]etry", },
-		{ .key = 'i', .descr = "[i]gnore", },
-		{ .key = 'I', .descr = "[I]gnore for all", },
+		{ .key = 'r', .descr = "[r]etry/", },
+		{ .key = 'i', .descr = "[i]gnore/", },
+		{ .key = 'I', .descr = "[I]gnore for all/", },
 		{ .key = 'a', .descr = "[a]bort", },
 		{ },
 	};
@@ -1168,7 +1248,14 @@ prompt_user(const io_args_t *args, const char title[], const char msg[],
 	/* Active cancellation conflicts with input processing by putting terminal in
 	 * a cooked mode. */
 	ui_cancellation_push_off();
-	const char response = curr_ops->choose(title, msg, variants);
+
+	const custom_prompt_t prompt = {
+		.title = title,
+		.message = msg,
+		.variants = variants,
+	};
+	char response = curr_ops->choose(&prompt);
+
 	ui_cancellation_pop();
 
 	return response;
@@ -1193,7 +1280,7 @@ run_operation_command(ops_t *ops, char cmd[], int cancellable)
 		return result_from_code(bg_and_wait_for_errors(cmd, &no_cancellation));
 	}
 
-	if(ops != NULL && ops->bg)
+	if(ops_runs_in_bg(ops))
 	{
 		const cancellation_t bg_cancellation_info = {
 			.arg = ops->bg_op,
@@ -1215,6 +1302,14 @@ run_operation_command(ops_t *ops, char cmd[], int cancellable)
 
 #endif
 
+/* Checks whether operation is a background one.  The parameter can be NULL.
+ * Returns non-zero for background operations. */
+static int
+ops_runs_in_bg(const ops_t *ops)
+{
+	return (ops != NULL && ops->bg);
+}
+
 /* Implementation of cancellation hook for background tasks. */
 static int
 bg_cancellation_hook(void *arg)
