@@ -308,7 +308,7 @@ TEST(good_flag_macros)
 
 	expanded = ma_expand("%N echo log", "", &flags, MER_OP);
 	assert_string_equal(" echo log", expanded);
-	assert_int_equal(MF_KEEP_SESSION, flags);
+	assert_int_equal(MF_KEEP_IN_FG, flags);
 	free(expanded);
 
 	expanded = ma_expand("%Pl echo log", "", &flags, MER_OP);
@@ -640,7 +640,7 @@ TEST(dollar_and_backtick_are_escaped_in_dquotes)
 	lwin.list_pos = 3;
 	assert_success(replace_string(&lwin.dir_entry[lwin.list_pos].name, "a$`b"));
 
-	curr_stats.shell_type = ST_NORMAL;
+	curr_stats.shell_type = ST_POSIX;
 	expanded = ma_expand("%\"c", "", NULL, MER_OP);
 	assert_string_equal("\"a\\$\\`b\"", expanded);
 	free(expanded);
@@ -651,7 +651,7 @@ TEST(dollar_and_backtick_are_escaped_in_dquotes)
 	free(expanded);
 
 	/* Restore normal value or some further tests can get broken. */
-	curr_stats.shell_type = ST_NORMAL;
+	curr_stats.shell_type = ST_POSIX;
 }
 
 TEST(newline_is_escaped_with_quotes)
@@ -749,7 +749,7 @@ TEST(flags_to_str)
 	assert_string_equal("%i", ma_flags_to_str(MF_IGNORE));
 	assert_string_equal("%n", ma_flags_to_str(MF_NO_TERM_MUX));
 
-	assert_string_equal("%N", ma_flags_to_str(MF_KEEP_SESSION));
+	assert_string_equal("%N", ma_flags_to_str(MF_KEEP_IN_FG));
 
 	assert_string_equal("%Pl", ma_flags_to_str(MF_PIPE_FILE_LIST));
 	assert_string_equal("%Pz", ma_flags_to_str(MF_PIPE_FILE_LIST_Z));
